package PScm;

use strict;
use warnings;
use PScm::Read;
use PScm::Env;
use PScm::Primitive;
use PScm::SpecialForm;
use PScm::Class;
use PScm::Continuation;
use FileHandle;
use Carp;

require Exporter;

our @ISA    = qw(Exporter);
our @EXPORT = qw(ReadEvalPrint);

=head1 NAME

PScm - PScm interpreter written in Perl

=head1 SYNOPSIS

  use PScm;
  ReadEvalPrint($in_filehandle[, $out_filehandle]);

=head1 DESCRIPTION

Just messing about, A toy lisp interpreter.

=cut

my @thread_queue;

sub ReadEvalPrint {
    my ($infh, $outfh) = @_;

    $outfh ||= new FileHandle(">-");
    my $reader      = new PScm::Read($infh);
    my $initial_env;
    $initial_env = new PScm::Env(
        let          => new PScm::SpecialForm::Let(),
        '*'          => new PScm::Primitive::Multiply(),
        '-'          => new PScm::Primitive::Subtract(),
        '+'          => new PScm::Primitive::Add(),
        if           => new PScm::SpecialForm::If(),
        lambda       => new PScm::SpecialForm::Lambda(),
        list         => new PScm::Primitive::List(),
        car          => new PScm::Primitive::Car(),
        cdr          => new PScm::Primitive::Cdr(),
        cons         => new PScm::Primitive::Cons(),
        letrec       => new PScm::SpecialForm::LetRec(),
        'let*'       => new PScm::SpecialForm::LetStar(),
        eval         => new PScm::SpecialForm::Eval(),
        macro        => new PScm::SpecialForm::Macro(),
        quote        => new PScm::SpecialForm::Quote(),
        'set!'       => new PScm::SpecialForm::Set(),
        begin        => new PScm::SpecialForm::Begin(),
        define       => new PScm::SpecialForm::Define(),
        'make-class' => new PScm::SpecialForm::MakeClass(),
        'call/cc'    => new PScm::SpecialForm::CallCC(),
        print        => new PScm::SpecialForm::Print($outfh),
        spawn        => new PScm::SpecialForm::Spawn(),
        exit         => new PScm::SpecialForm::Exit(),
        error        => new PScm::SpecialForm::Error(
                        $outfh,
                        bounce { repl($initial_env, $reader, $outfh) }
                     ),
        amb          => new PScm::SpecialForm::Amb(),
        'eq?'        => new PScm::Primitive::Eq(),
        '>'          => new PScm::Primitive::Gt(),
        '<'          => new PScm::Primitive::Lt(),
        '>='         => new PScm::Primitive::Ge(),
        '<='         => new PScm::Primitive::Le(),
        and          => new PScm::SpecialForm::And(),
        or           => new PScm::SpecialForm::Or(),
        unify        => new PScm::SpecialForm::Unify(),
        substitute   => new PScm::Primitive::Substitute(),
        'new-env'    => new PScm::Primitive::NewEnv(),
        instantiate  => new PScm::Primitive::Instantiate(),
        'pair?'      => new PScm::Primitive::TypeCheck::Pair(),
        'var?'       => new PScm::Primitive::TypeCheck::Var(),
        'number?'    => new PScm::Primitive::TypeCheck::Number(),
    );

    $initial_env->Define(
        PScm::Expr::Symbol->new("root"),
        PScm::Class::Root->new($initial_env)
    );
    __PACKAGE__->new_thread(bounce { repl($initial_env, $reader, $outfh) });
    trampoline();
}

sub repl {
    my ($env, $reader, $outfh, $fail1) = @_;
    $fail1 ||= fail { __PACKAGE__->Error("no current problem", $env) };
    $reader->Read(
        cont {
            my ($expr, $fail2) = @_;
            $expr->Eval(
                $env,
                cont {
                    my ($result, $fail3) = @_;
                    $result->Print(
                        $outfh,
                        cont {
                            my ($dummy, $fail4) = @_;
                            repl($env, $reader, $outfh, $fail4);
                        },
                        $fail3
                    )
                },
                fail {
                    __PACKAGE__->Error("no more solutions", $env);
                }
            )
        },
        $fail1
    )
}

sub new_thread {
    my ($self, $bounce) = @_;
    push @thread_queue, $bounce;
}

sub trampoline {
    while (@thread_queue) {
        my $bounce = shift @thread_queue;
        $bounce = $bounce->Bounce();
        push @thread_queue, $bounce if defined $bounce;
    }
}

sub Print {
    my ($expr, $outfh, $cont, $fail) = @_;
    print $outfh $expr->as_string, "\n";
    $cont->Cont($expr, $fail);
}

sub Error {
    my ($self, $msg, $env) = @_;
    my $error = $env->LookUp(new PScm::Expr::Symbol('error'));
    $error->do_error($msg);
}

sub as_string { ref($_[0]); }

sub new { bless {}, $_[0] }

sub checkTypes {
    my ($self, $args, $types) = @_;
    my @args = @$args;
    my @types = @$types;
    while (@types) {
        my $type = shift @types;
        my $arg = shift @args;
        confess "not enough arguments" unless defined $arg;
        confess "$arg is not a $type" unless $arg->isa("PScm::$type");
    }
}

1;
