package PScm::Env;

use strict;
use warnings;
use base qw(PScm::Expr);
use PScm::Continuation;

sub new {
    my ($class, %bindings) = @_;

    $class = ref($class) || $class;
    bless { bindings => {%bindings}, parent => 0 }, $class;
}

sub Extend {
    my ($self, $symbols, $values, $cont, $fail) = @_;

    $values->map_eval(
        $self,
        cont {
            my ($evaluated_values, $fail) = @_;
            $self->ExtendUnevaluated(
                $symbols, $evaluated_values, $cont, $fail
            )
        },
        $fail
    );
}

sub ExtendUnevaluated {
    my ($self, $symbols, $values, $cont, $fail) = @_;

    my %bindings;
    eval {
        $self->_populate_bindings(\%bindings, $symbols, $values);
    };
    if ($@) {
        $self->Error($@, $self);
    } else {
        my $newenv = $self->new(%bindings);
        $newenv->{parent} = $self;
        $cont->Cont($newenv, $fail);
    }
}

sub _populate_bindings {
    my ($self, $bindings, $symbols, $values) = @_;
    if ($symbols->is_null) {
        die "too many arguments\n" unless $values->is_null;
    } elsif ($symbols->is_pair) {
        if ($values->is_pair) {
            my $symbol = $symbols->first;
            if ($symbol->is_symbol) {
                $bindings->{$symbol->value} = $values->first;
                $self->_populate_bindings($bindings,
                                          $symbols->rest,
                                          $values->rest);
            } else {
                die "bad formal arguments[1]:",
                    $symbol->as_string(), "\n";
            }
        } else {
            die "not enough arguments\n";
        }
    } elsif ($symbols->is_symbol) {
        $bindings->{$symbols->value} = $values;
    } else {
        die "bad formal arguments[2]:",
            $symbols->as_string(), "\n";
    }
}

sub ExtendRecursively {
    my ($self, $symbols, $values, $cont, $fail) = @_;

    $self->ExtendUnevaluated(
        $symbols,
        $values,
        cont {
            my ($newenv, $fail) = @_;
            $newenv->_eval_values($cont, $fail);
        },
        $fail
    );
}

sub ExtendIteratively {
    my ($self, $symbols, $values, $cont, $fail) = @_;
    if ($symbols->isTrue) {
        $self->Extend(
            new PScm::Expr::List($symbols->first),
            new PScm::Expr::List($values->first),
            cont {
                my ($extended, $fail) = @_;
                $extended->ExtendIteratively($symbols->rest,
                    $values->rest, $cont, $fail);
            },
            $fail
        );
    } else {
        $cont->Cont($self, $fail);
    }
}

sub _eval_values {
    my ($self, $cont, $fail) = @_;
    $self->_map_bindings([keys %{ $self->{bindings} }], $cont, $fail);
}

sub _map_bindings {
    my ($self, $ra_keys, $cont, $fail) = @_;
    my (@keys) = @$ra_keys;
    if (@keys) {
        my $firstkey = shift @keys;
        $self->{bindings}{$firstkey}->Eval(
            $self,
            cont {
                my ($value, $fail) = @_;
                $self->{bindings}{$firstkey} = $value;
                $self->_map_bindings([@keys], $cont, $fail);
            },
            $fail
        );
    } else {
        $cont->Cont($self, $fail);
    }
}

sub LookUp {
    my ($self, $symbol) = @_;

    if (defined(my $ref = $self->_lookup_ref($symbol))) {
        return $$ref;
    } else {
        die "no binding for @{[$symbol->value]}",
          " in @{[ref($self)]}\n";
    }
}

sub LookUpNoError {
    my ($self, $symbol) = @_;

    if (defined(my $ref = $self->_lookup_ref($symbol))) {
        return $$ref;
    } else {
        return undef;
    }
}

sub LookUpHere {
    my ($self, $symbol) = @_;
    if (exists($self->{bindings}{ $symbol->value })) {
        return $self->{bindings}{ $symbol->value };
    } else {
        return undef;
    }
}

sub Assign {
    my ($self, $symbol, $value) = @_;

    if (defined(my $ref = $self->_lookup_ref($symbol))) {
        $$ref = $value;
    } else {
        die "no binding for @{[$symbol->value]}",
          " in @{[ref($self)]}\n";
    }
}

sub _lookup_ref {
    my ($self, $symbol) = @_;

    if (exists($self->{bindings}{ $symbol->value })) {
        return \$self->{bindings}{ $symbol->value };
    } elsif ($self->{parent}) {
        return $self->{parent}->_lookup_ref($symbol);
    } else {
        return undef;
    }
}

sub UnSet {
    my ($self, $symbol) = @_;
    delete $self->{bindings}{ $symbol->value };
}

sub _lookup_method {
    my ($self, $method_name) = @_;

    return $self->_lookup_method_here($method_name)
      || $self->_lookup_method_in_super($method_name);
}

sub _lookup_method_here {
    my ($self, $method_name) = @_;

    if (exists $self->{bindings}{class}) {
        return $self->{bindings}{class}
          ->get_method($method_name, $self);
    }
}

sub _lookup_method_in_super {
    my ($self, $method_name) = @_;

    if (exists $self->{bindings}{super}) {
        return $self->{bindings}{super}
          ->_lookup_method($method_name);
    }
}

sub Define {
    my ($self, $symbol, $value) = @_;

    $self->{bindings}{ $symbol->value } = $value;
    return $symbol;
}

sub call_method {
    my ($self, $this, $method_name, $args, $env, $cont, $fail) = @_;

    if (my $method = $self->_lookup_method($method_name)) {
        $method->ApplyMethod($this, $args, $env, $cont, $fail);
    } else {
        $cont->Cont(undef, $fail);
    }
}

sub Apply {
    my ($self, $form, $env, $cont, $fail) = @_;

    my ($method, $args) = ($form->first, $form->rest);
    $self->CallMethodOrDie($self, $method, $args, $env, $cont, $fail);
}

sub CallMethodOrDie {
    my ($self, $this, $method, $args, $env, $cont, $fail) = @_;
    $self->call_method(
        $this,
        $method->value,
        $args, $env,
        cont {
            my ($res, $fail) = @_;
            if (defined $res) {
                $cont->Cont($res, $fail);
            } else {
                $self->Error(
                    "method " . $method->value . " not found\n",
                    $env
                );
            }
        },
        $fail
    );
}

sub ResolveAll {
    my ($self) = @_;
    my %bindings;
    foreach my $var ($self->Keys) {
        $bindings{$var} = $self->Resolve(new PScm::Expr::Var($var));
    }
    return $self->new(%bindings);
}

sub Keys {
    my ($self, $seen) = @_;
    $seen ||= {};
    foreach my $key (keys %{$self->{bindings}}) {
        $seen->{$key} = 1;
    }
    if ($self->{parent}) {
        $self->{parent}->Keys($seen);
    }
    return (keys %$seen);
}

sub Resolve {
    my ($self, $term) = @_;
    while ($term->is_var) {
        if (my $val = $self->LookUpNoError($term)) {
            $term = $val;
        } else {
            return $term;
        }
    }
    return $term->ResolveTerm($self);
}


#########################
package PScm::Env::Super;

use base qw(PScm::Env);

sub Apply {
    my ($self, $form, $env, $cont, $fail) = @_;

    my ($method, $args) = ($form->first, $form->rest);
    my $this = $env->LookUp(PScm::Expr::Symbol->new("this"));
    $self->CallMethodOrDie($this, $method, $args, $env, $cont, $fail);
}

1;
