package PScm::Primitive;

use strict;
use warnings;
use base qw(PScm::Expr);
use PScm::Continuation;

sub Apply {
    my ($self, $form, $env, $cont, $fail) = @_;

    $form->map_eval(
        $env,
        cont {
            my ($evaluated_args, $fail) = @_;
            my $result = eval {
                $self->_apply($evaluated_args->value);
            };
            if ($@) {
                $self->Error($@, $env);
            } else {
                $cont->Cont($result, $fail);
            }
        },
        $fail
    );
}

sub _check_type {
    my ($self, $thing, $type) = @_;

    die "wrong type argument(", ref($thing), ") to ", ref($self),
      "\n"
      unless $thing->isa($type);
}

##################################
package PScm::Primitive::Multiply;

use base qw(PScm::Primitive);

sub _apply {
    my ($self, @args) = @_;

    my $result = PScm::Expr::Number->new(1)->value();

    while (@args) {
        my $arg = shift @args;
        $self->_check_type($arg, 'PScm::Expr::Number');
        $result *= $arg->value;
    }

    return new PScm::Expr::Number($result);
}

##################################
package PScm::Primitive::Subtract;

use base qw(PScm::Primitive);

sub _apply {
    my ($self, @args) = @_;

    unshift @args, PScm::Expr::Number->new(0) if @args < 2;

    my $arg = shift @args;
    $self->_check_type($arg, 'PScm::Expr::Number');

    my $result = $arg->value;

    while (@args) {
        $arg = shift @args;
        $self->_check_type($arg, 'PScm::Expr::Number');
        $result -= $arg->value;
    }

    return new PScm::Expr::Number($result);
}

##################################
package PScm::Primitive::Add;

use base qw(PScm::Primitive);

sub _apply {
    my ($self, @args) = @_;

    my $result = 0;

    while (@args) {
        my $arg = shift @args;
        $self->_check_type($arg, 'PScm::Expr::Number');
        $result += $arg->value;
    }

    return new PScm::Expr::Number($result);
}

##############################
package PScm::Primitive::List;

use base qw(PScm::Primitive);

sub _apply {
    my ($self, @args) = @_;

    return new PScm::Expr::List(@args);
}

#############################
package PScm::Primitive::Car;

use base qw(PScm::Primitive);

sub _apply {
    my ($self, $arg) = @_;

    $self->_check_type($arg, 'PScm::Expr::List');
    return $arg->first;
}

#############################
package PScm::Primitive::Cdr;

use base qw(PScm::Primitive);

sub _apply {
    my ($self, $arg) = @_;

    $self->_check_type($arg, 'PScm::Expr::List');
    return $arg->rest;
}

##############################
package PScm::Primitive::Cons;

use base qw(PScm::Primitive);

sub _apply {
    my ($self, $car, $cdr) = @_;

    return PScm::Expr::List->Cons($car, $cdr);
}

############################
package PScm::Primitive::Eq;

use base qw(PScm::Primitive);

sub _apply {
    my ($self, @things) = @_;
    while (@things > 1) {
        my $thing = shift @things;
        return PScm::Expr::Number->new(0) unless $thing->Eq($things[0]);
    }
    return PScm::Expr::Number->new(1);
}

#################################
package PScm::Primitive::Compare;

use base qw(PScm::Primitive);

sub _apply {
    my ($self, @numbers) = @_;
    $self->_check_type($numbers[0], 'PScm::Expr::Number') if @numbers;
    while (@numbers > 1) {
        my $number = shift @numbers;
        $self->_check_type($numbers[0], 'PScm::Expr::Number');
        return PScm::Expr::Number->new(0)
            unless $self->_compare($number, $numbers[0]);
    }
    return PScm::Expr::Number->new(1);
}

############################
package PScm::Primitive::Lt;

use base qw(PScm::Primitive::Compare);

sub _compare {
    my ($self, $first, $second) = @_;
    return $first->value < $second->value;
}

############################
package PScm::Primitive::Gt;

use base qw(PScm::Primitive::Compare);

sub _compare {
    my ($self, $first, $second) = @_;
    return $first->value > $second->value;
}

############################
package PScm::Primitive::Le;

use base qw(PScm::Primitive::Compare);

sub _compare {
    my ($self, $first, $second) = @_;
    return $first->value <= $second->value;
}

############################
package PScm::Primitive::Ge;

use base qw(PScm::Primitive::Compare);

sub _compare {
    my ($self, $first, $second) = @_;
    return $first->value >= $second->value;
}

####################################
package PScm::Primitive::Substitute;

use base qw(PScm::Primitive);

sub _apply {
    my ($self, $body, $qenv) = @_;
    $body->Substitute($qenv->ResolveAll());
}

###################################
package PScm::Primitive::TypeCheck;

use base qw(PScm::Primitive);

sub _apply {
    my ($self, $body) = @_;
    if ($self->test($body)) {
        return new PScm::Expr::Number(1);
    } else {
        return new PScm::Expr::Number(0);
    }
}

#########################################
package PScm::Primitive::TypeCheck::Pair;
use base qw(PScm::Primitive::TypeCheck);

sub test { $_[1]->is_pair }

########################################
package PScm::Primitive::TypeCheck::Var;
use base qw(PScm::Primitive::TypeCheck);

sub test { $_[1]->is_var }

###########################################
package PScm::Primitive::TypeCheck::Number;
use base qw(PScm::Primitive::TypeCheck);

sub test { $_[1]->is_number }

################################
package PScm::Primitive::NewEnv;

use base qw(PScm::Primitive);

sub _apply {
    new PScm::Env();
}

#####################################
package PScm::Primitive::Instantiate;

use base qw(PScm::Primitive);

sub new {
    my ($class) = @_;
    bless {
        seen => {},
        counter => 0,
    }, $class;
}

sub _apply {
    my ($self, $body) = @_;
    $self->{seen} = {};
    $body->Instantiate($self);
}

sub Replace {
    my ($self, $var) = @_;
    return $var if $var->is_anon;
    unless (exists($self->{seen}{$var->value})) {
        $self->{seen}{$var->value} =
            new PScm::Expr::Var($self->{counter}++);
    }
    return $self->{seen}{$var->value};
}

1;
