use strict;
use warnings;
use Test::More;
use lib './t/lib';
use PScm::Test tests => 13;

BEGIN { use_ok('PScm') }

my $prereqs = <<EOT;

(define not
  (lambda (x)
     (if x 0 1)))

(define require
  (lambda (x)
    (if x x (amb))))

(define one-of
  (lambda (lst)
    (begin
        (require lst)
        (amb (car lst) (one-of (cdr lst))))))

(define no-vars?
  (lambda (expr)
    (if (pair? expr)
        (and (no-vars? (car expr))
             (no-vars? (cdr expr)))
        (not (var? expr)))))

(define substitute-all
  (lambda (expr env)
    (let ((subst-expr (substitute expr env)))
         (begin
           (require (no-vars? subst-expr))
           subst-expr))))

(define prove
  (lambda (goals)
    (substitute-all goals
                    (match-goals goals
                                 (new-env)))))

(define match-goals
  (lambda (goals env)
    (if goals
        (match-goals (cdr goals)
                     (match-goal (car goals)
                                 env))
        env)))

(define match-goal
  (lambda (goal env)
    (if (eq? (car (cdr goal)) 'is)
        (match-is goal env)
        (if (eq? (car goal) 'require)
            (match-require goal env)
            (match-goal-to-rule goal
                                (one-of the-rules)
                                env)))))

(define match-is
  (lambda (goal env)
    (let* ((var (car goal))
           (value (car (cdr (cdr goal))))
           (svalue (substitute-all value env)))
          (unify var (eval svalue) env))))

(define match-require
(lambda (goal env)
(let ((sgoal (substitute-all goal env)))
  (begin
    (eval sgoal)
    env))))

(define match-goal-to-rule
  (lambda (goal rule env)
    (let* ((instantiated-rule (instantiate rule))
           (head (car instantiated-rule))
           (body (cdr instantiated-rule))
           (extended-env (unify (substitute goal env)
                                head
                                env)))
      (match-goals body extended-env))))
EOT

my $prereqs_output = <<EOT;
not
require
one-of
no-vars?
substitute-all
prove
match-goals
match-goal
match-is
match-require
match-goal-to-rule
EOT

$prereqs_output =~ s/\n$//s;


eval_ok(<<EOT, <<EOR, 'socrates');
$prereqs
(define the-rules
  (list '((man socrates))
        '((mortal X) (man X))))
(prove '((mortal socrates)))
EOT
$prereqs_output
the-rules
((mortal socrates))
EOR

my $rules = <<'EOT';
(define the-rules
  (list '((mary likes cheese))
        '((mary likes wine))
        '((john likes beer))
        '((john likes wine))
        '((john likes chips))
        '((person mary))
        '((person john))
        '((mary likes X) (person X)
                         (require (not (eq? 'X 'mary)))
                         (X likes Y)
                         (mary likes Y))))
EOT

eval_ok(<<EOT, <<EOR, 'mary and john [1]');
$prereqs
$rules
(prove '((mary likes john)))
EOT
$prereqs_output
the-rules
((mary likes john))
EOR

eval_ok(<<EOT, <<EOR, 'mary and john [2]');
$prereqs
$rules
(prove '((mary likes X)))
?
?
?
EOT
$prereqs_output
the-rules
((mary likes cheese))
((mary likes wine))
((mary likes john))
Error: no more solutions
EOR

$rules = <<'EOT';
(define the-rules
  (list '((append () Y Y))
        '((append (A . X) Y (A . Z)) (append X Y Z))))
EOT

eval_ok(<<EOT, <<EOR, 'append [1]');
$prereqs
$rules
(prove '((append (a b) (c d) X)))
EOT
$prereqs_output
the-rules
((append (a b) (c d) (a b c d)))
EOR

eval_ok(<<EOT, <<EOR, 'append [2]');
$prereqs
$rules
(prove '((append X Y (a b c d))))
?
?
?
?
?
EOT
$prereqs_output
the-rules
((append () (a b c d) (a b c d)))
((append (a) (b c d) (a b c d)))
((append (a b) (c d) (a b c d)))
((append (a b c) (d) (a b c d)))
((append (a b c d) () (a b c d)))
Error: no more solutions
EOR

eval_ok(<<EOT, <<EOR, 'symbolic differentiation');
$prereqs
(define the-rules
  (list '((derivative X X 1))
        '((derivative N X 0) (require (number? 'N)))
        '((derivative (X ^ N) X (N * (X ^ P))) (P is (- N 1)))
        '((derivative (log X) X (1 / X)))
        '((derivative (F + G) X (DF + DG))
            (derivative F X DF)
            (derivative G X DG))
        '((derivative (F - G) X (DF - DG))
            (derivative F X DF)
            (derivative G X DG))
        '((derivative (F * G) X ((F * DG) + (G * DF)))
            (derivative F X DF)
            (derivative G X DG))
        '((derivative (1 / F) X ((- DF) / (F * F)))
            (derivative F X DF))
        '((derivative (F / G) X (((G * DF) - (F * DG)) / (G * G)))
            (derivative F X DF)
            (derivative G X DG))))

(prove '((derivative (((x ^ 2) + x) + 1) x X)))
EOT
$prereqs_output
the-rules
((derivative (((x ^ 2) + x) + 1) x (((2 * (x ^ 1)) + 1) + 0)))
EOR

$rules = <<EOF;
(define the-rules
  (list '((add X Y Z) (Z is (+ X Y)))
        '((add X Y Z) (X is (- Z Y)))
        '((add X Y Z) (Y is (- Z X)))))
EOF

eval_ok(<<EOF, <<EOR, 'builtin arithmetic');
$prereqs
$rules
(prove '((add 2 3 X)))
(prove '((add 2 X 5)))
(prove '((add X 3 5)))
EOF
$prereqs_output
the-rules
((add 2 3 5))
((add 2 3 5))
((add 2 3 5))
EOR

$rules = <<EOF;
(define the-rules
  (list '(((johann ambrosius) father-of (j s)))
        '(((j c friedrich) father-of (w f ernst)))
        '(((j s) (anna magdalena) parents-of (j c friedrich)))
        '(((j s) (anna magdalena) parents-of (johann christian)))
        '(((j s) (maria barbara) parents-of (wilhelm friedmann)))
        '(((j s) (maria barbara) parents-of (c p e)))
        '((X father-of Y) (X _ parents-of Y))
        '((X mother-of Y) (_ X parents-of Y))
        '((X parent-of Y) (X father-of Y))
        '((X parent-of Y) (X mother-of Y))
        '((X ancestor-of Y) (X parent-of Y))
        '((X ancestor-of Z) (X parent-of Y) (Y ancestor-of Z))))
EOF

# had to truncate this test - takes too long to run
eval_ok(<<EOF, <<EOR, 'genealogy');
$prereqs
$rules
(prove '((X ancestor-of (w f ernst))))
EOF
$prereqs_output
the-rules
(((j c friedrich) ancestor-of (w f ernst)))
EOR

$rules = <<EOF;
(define the-rules
  (list
        '((proper-noun (john . S) S))
        '((proper-noun (paul . S) S))
        '((noun (car . S) S))
        '((noun (garage . S) S))
        '((auxilliary (will . S) S))
        '((auxilliary (has . S) S))
        '((verb (put . S) S))
        '((article (the . S) S))
        '((article (a . S) S))
        '((article (his . S) S))
        '((preposition (in . S) S))
        '((preposition (to . S) S))
        '((preposition (with . S) S))
        '((degree (very . S) S))
        '((degree (quite . S) S))
        '((adjective (red . S) S))
        '((adjective (green . S) S))
        '((adjective (old . S) S))
        '((adjective (new . S) S))
        '((sentance S X)
            (noun-phrase S S1) (verb-phrase S1 X))
        '((sentance S X)
            (noun-phrase S S1) (auxilliary S1 S2) (verb-phrase S2 X))
        '((noun-phrase S X)
            (article S S1) (adj-phrase S1 X))
        '((noun-phrase S X) (proper-noun S X))
        '((adj-phrase S X) (noun S X))
        '((adj-phrase S X)
            (degree-phrase S S1) (adj-phrase S1 X))
        '((degree-phrase S X) (adjective S X))
        '((deg-phrase S X)
            (degree S S1) (deg-phrase S1 X))
        '((verb-phrase S X)
            (verb S S1) (noun-phrase S1 S2) (prep-phrase S2 X))
        '((prep-phrase S X)
            (preposition S S1) (noun-phrase S1 X))

    ))
EOF

eval_ok(<<EOF, <<EOT, 'parsing');
$prereqs
$rules
(prove '((sentance (john will put his car in the garage) ())))
EOF
$prereqs_output
the-rules
((sentance (john will put his car in the garage) ()))
EOT

$rules = <<EOF;
(define the-rules
  (list
    '((simplify E E) (require (not (pair? 'E))))
    '((simplify (X OP Y) E)
        (simplify X X1)
        (simplify Y Y1)
        (s (X1 OP Y1) E))
    '((s (X + 0) X))
    '((s (0 + X) X))
    '((s (X + Y) Z)
        (require (and (number? 'X) (number? 'Y)))
        (Z is (+ X Y)))
    '((s (X + Y) (X + Y)))
    '((s (X * 0) 0))
    '((s (0 * X) 0))
    '((s (X * 1) X))
    '((s (1 * X) X))
    '((s (X * Y) Z)
        (require (and (number? 'X) (number? 'Y)))
        (Z is (* X Y)))
    '((s (X * Y) (X * Y)))
    '((s (X ^ 0) 1))
    '((s (X ^ 1) X))
    '((s (X ^ Y) (X ^ Y)))))
EOF

eval_ok(<<EOF, <<EOT, 'simplification');
$prereqs
$rules
(prove '((simplify (((2 * (x ^ 1)) + 1) + 0) X)))
EOF
$prereqs_output
the-rules
((simplify (((2 * (x ^ 1)) + 1) + 0) ((2 * x) + 1)))
EOT

$rules = <<EOR;
(define the-rules
  (list
     '((factorial 0 1))
     '((factorial N X) (T is (- N 1)) (factorial T U) (X is (* N U)))))
EOR

eval_ok(<<EOF, <<EOT, 'recursion');
$prereqs
$rules
(prove '((factorial 10 X)))
EOF
$prereqs_output
the-rules
((factorial 10 3628800))
EOT

eval_ok(<<EOF, <<EOT, 'factorial not bidirectional');
$prereqs
$rules
(prove '((factorial X 3628800)))
EOF
$prereqs_output
the-rules
Error: no more solutions
EOT

# vim: ft=perl
