use strict;
use warnings;
use Test::More;
use lib './t/lib';
use PScm::Test tests => 8;

BEGIN { use_ok('PScm') }

my $prereqs = <<EOT;
(define require
  (lambda (x)
    (if x x (amb))))

(define member?
  (lambda (item lst)
    (if lst
        (if (eq? item (car lst))
            1
            (member? item (cdr lst)))
        0)))

(define distinct?
  (lambda (lst)
    (if lst
        (if (member? (car lst)
                     (cdr lst))
            0
            (distinct? (cdr lst)))
        1)))

(define one-of
  (lambda (lst)
    (begin
        (require lst)
        (amb (car lst) (one-of (cdr lst))))))

(define exclude
  (lambda (items lst)
    (if lst
        (if (member? (car lst) items)
            (exclude items (cdr lst))
            (cons (car lst)
                  (exclude items (cdr lst))))
        ())))

(define abs
  (lambda (x)
    (if (< x 0)
        (- x)
        x)))

(define not
  (lambda (x)
    (if x 0 1)))

(define xor
  (lambda (x y)
    (or (and x (not y))
        (and y (not x)))))

(define difference
  (lambda (a b)
    (abs (- a b))))

(define divisible-by
  (lambda (n)
    (lambda (v)
      (begin
        (define test
          (lambda (o)
            (if (eq? o v)
                1
                (if (> o v)
                    0
                    (test (+ o n))))))
        (test 0)))))

(define even?
  (lambda (a)
    ((divisible-by 2) a)))
EOT

my $prereqs_output = <<EOT;
require
member?
distinct?
one-of
exclude
abs
not
xor
difference
divisible-by
even?
EOT

$prereqs_output =~ s/\n$//s;


eval_ok(<<EOT, <<EOR, 'even?');
$prereqs
(define test
  (lambda ()
    (let ((x (amb 1 2 3 4 5)))
      (begin
        (require (even? x))
        x))))
(test)
?
?
EOT
$prereqs_output
test
2
4
Error: no more solutions
EOR

eval_ok(<<EOT, <<EOR, 'Barrels of Fun');
$prereqs
(define some-of
  (lambda (lst)
    (begin
      (require lst)
      (amb (list (car lst))
           (some-of (cdr lst))
           (cons (car lst)
                 (some-of (cdr lst)))))))

(define sum
  (lambda (lst)
    (if lst
      (+ (car lst)
         (sum (cdr lst)))
      0)))

(define barrels-of-fun
  (lambda ()
    (let* ((barrels (list 30 32 36 38 40 62))
           (beer (one-of barrels))
           (wine (exclude (list beer) barrels))
           (barrel1 (one-of wine))
           (barrel2 (one-of (exclude (list barrel1) wine)))
           (barrels (some-of (exclude (list barrel1 barrel2) wine))))
      (begin
        (require (eq? (* 2 (+ barrel1 barrel2))
                      (sum barrels)))
        beer))))
(barrels-of-fun)
EOT
$prereqs_output
some-of
sum
barrels-of-fun
40
EOR

# Baker, Cooper, Fletcher, Miller and Smith live on different
# floors of a five-storey building. Baker does not live on the
# top floor. Cooper does not live on the bottom floor. Fletcher
# does not live on the top or the bottom floor. Miller lives
# on a higher floor than Cooper. Smith does not live on a
# floor adjacent to Fletcher's. Fletcher does not live on a floor
# adjacent to Cooper's. Where does everyone live?

eval_ok(<<EOT, <<EOR, 'amb');
$prereqs

(define multiple-dwelling
  (lambda ()
    (let* ((baker (one-of (list 1 2 3 4)))
           (cooper (one-of (exclude (list baker) (list 2 3 4 5))))
           (fletcher (one-of (exclude (list baker cooper) (list 2 3 4))))
           (miller (one-of (exclude (list baker cooper fletcher)
                                    (list 1 2 3 4 5))))
           (smith (car (exclude (list baker cooper fletcher miller)
                                (list 1 2 3 4 5)))))
       (begin
         (require (> miller cooper))
         (require (not (eq? (difference smith fletcher) 1)))
         (require (not (eq? (difference cooper fletcher) 1)))
         (list (list 'baker baker)
               (list 'cooper cooper)
               (list 'fletcher fletcher)
               (list 'miller miller)
               (list 'smith smith))))))

(multiple-dwelling)
EOT
$prereqs_output
multiple-dwelling
((baker 3) (cooper 2) (fletcher 4) (miller 5) (smith 1))
EOR

eval_ok(<<EOF, <<EOR, 'Liars (optimized)');
$prereqs
(define liars
  (lambda ()
    (let* ((betty (amb 1 2 3 4 5))
           (ethel (one-of (exclude (list betty)
                                   (list 1 2 3 4 5))))
           (joan (one-of (exclude (list betty ethel)
                                  (list 1 2 3 4 5))))
           (kitty (one-of (exclude (list betty ethel joan)
                                   (list 1 2 3 4 5))))
           (mary (car (exclude (list betty ethel joan kitty)
                               (list 1 2 3 4 5)))))
         (begin
           (require (xor (eq? kitty 2) (eq? betty 3)))
           (require (xor (eq? ethel 1) (eq? joan 2)))
           (require (xor (eq? joan 3) (eq? ethel 5)))
           (require (xor (eq? kitty 2) (eq? mary 4)))
           (require (xor (eq? mary 4) (eq? betty 1)))
           '((betty ,betty)
             (ethel ,ethel)
             (joan ,joan)
             (kitty ,kitty)
             (mary ,mary))))))
(liars)
?
EOF
$prereqs_output
liars
((betty 3) (ethel 5) (joan 2) (kitty 1) (mary 4))
Error: no more solutions
EOR

eval_ok(<<EOF, <<EOR, 'set! backtracking');
(let ((x 1))
     (let ((y (amb 'a 'b)))
          (begin
            (print (list 'x x))
            (set! x 2)
            (print (list 'x x))
            y)))
?
EOF
(x 1)
(x 2)
a
(x 1)
(x 2)
b
EOR

eval_ok(<<EOF, <<EOR, 'define backtracking');
(define x 1)
(let ((y (amb 'a 'b)))
     (begin
       (print (list 'x x))
       (define x 2)
       (print (list 'x x))
       y))
?
EOF
x
(x 1)
(x 2)
a
(x 1)
(x 2)
b
EOR

eval_ok(<<EOT, <<EOR, 'parsing');
$prereqs

(define proper-nouns '(john paul))
(define nouns '(car garage))
(define auxilliaries '(will has))
(define verbs '(put))
(define articles '(the a his))
(define prepositions '(in to with))
(define degrees '(very quite))
(define adjectives '(red green old new))

(define parse-sentance
  (lambda ()
    (amb (list (parse-noun-phrase)
               (parse-word auxilliaries)
               (parse-verb-phrase))
         (list (parse-noun-phrase)
               (parse-verb-phrase)))))

(define parse-noun-phrase
  (lambda ()
    (amb (parse-word proper-nouns)
         (list (parse-word articles)
               (parse-adj-phrase)))))

(define parse-adj-phrase
  (lambda ()
    (amb (list (parse-deg-phrase)
               (parse-adj-phrase))
         (parse-word nouns))))

(define parse-deg-phrase
  (lambda ()
    (amb (list (parse-word degrees)
               (parse-deg-phrase))
         (parse-word adjectives))))

(define parse-verb-phrase
  (lambda ()
    (list (parse-word verbs)
          (parse-noun-phrase)
          (parse-prep-phrase))))

(define parse-prep-phrase
  (lambda ()
    (list (parse-word prepositions)
          (parse-noun-phrase))))

(define parse-word
  (lambda (words)
    (begin
      (require *unparsed*)
      (require (member? (car *unparsed*) words))
      (let ((found-word (car *unparsed*)))
        (begin
          (set! *unparsed* (cdr *unparsed*))
          found-word)))))

(define *unparsed* ())

(define parse
  (lambda (input)
    (begin
      (set! *unparsed* input)
      (let ((sentance (parse-sentance)))
        (begin
          (require (not *unparsed*))
          sentance)))))

(parse '(john will put his car in the garage))
(parse '(paul put a car in his garage))
(parse '(paul has put a very very old car in his quite new red garage))
EOT
$prereqs_output
proper-nouns
nouns
auxilliaries
verbs
articles
prepositions
degrees
adjectives
parse-sentance
parse-noun-phrase
parse-adj-phrase
parse-deg-phrase
parse-verb-phrase
parse-prep-phrase
parse-word
*unparsed*
parse
(john will (put (his car) (in (the garage))))
(paul (put (a car) (in (his garage))))
(paul has (put (a ((very (very old)) car)) (in (his ((quite new) (red garage))))))
EOR

# vim: ft=perl
